import json

def load_json(file_path):
    with open(file_path, 'r', encoding='utf-8') as fp:
        s = fp.read()
        data = json.loads(s)
    return data

def load_jsonl(url):
    data = []
    with open(url, 'r', encoding='utf-8') as f:
        for i in f:
            json_data = json.loads(i)
            data.append(json_data)
    return data

def write_json(data, file_path):
    output = json.dumps(data, indent=4, ensure_ascii=False)
    with open(file_path, "w+", encoding="utf-8") as file:
        file.write(output)
    return True



import os
import shutil
from natsort import natsorted
from urllib.parse import unquote
from urllib.parse import quote

def create_folder(path):
    if not os.path.exists(path):
        os.makedirs(path)
        return True
    else:
        return False
    
def get_file_names(directory):
    file_names = natsorted(os.listdir(directory))
    return file_names

def get_folder_names(directory):
    folder_names = natsorted(next(os.walk(directory))[1], key=str.lower)
    return folder_names

def rename_folder(old_name, new_name):
    os.rename(old_name, new_name)

def move_file(source_path, destination_path):
    try:
        shutil.move(source_path, destination_path)
    except Exception as e:
        print(f"移动失败: {e}")

def copy_any(source_path, destination_path):
    try:
        if os.path.isdir(source_path):
            shutil.copytree(source_path, destination_path)
        elif os.path.isfile(source_path):
            shutil.copy(source_path, destination_path)
        else:
            print("源路径无效：既不是文件也不是文件夹")
    except Exception as e:
        print(f"复制失败: {e}")

def rename_file(old_path, new_path):
    try:
        os.rename(old_path, new_path)
    except Exception as e:
        print(f"重命名失败: {e}")

def delete_file(file_path):
    try:
        os.remove(file_path)
    except Exception as e:
        print(f"删除失败: {e}")


import random
import keyboard
import os


def cmd_2048():
    mp = [['.' for _ in range(4)] for _ in range(4)]
    wall = [[0 for _ in range(4)] for _ in range(4)]

    while(1):
        xi = random.randint(0, 3)
        yi = random.randint(0, 3)
        if mp[xi][yi] == '.':
            yi = random.randint(0, 3)
            z = random.randint(0, 99)
            if z >= 90:
                mp[xi][yi] = 4
            else:
                mp[xi][yi] = 2
            break

    score = 0
    while(1):
        wall = [[0 for _ in range(4)] for _ in range(4)]
        
        while(1):
            xi = random.randint(0, 3)
            yi = random.randint(0, 3)
            if mp[xi][yi] == '.':
                z = random.randint(0, 99)
                if z >= 90:
                    mp[xi][yi] = 4
                else:
                    mp[xi][yi] = 2
                break

        for i in mp:
            for j in i:
                ed = (5 - len(str(j))) * " "
                print(j,end=ed)
            print()
        print("score:" + str(score))

        finished = 1
        for j in range(0,4):
            for i in range(1,4):
                if mp[i][j] != '.' and wall[i][j] == 0:
                    if mp[i - 1][j] == mp[i][j] and wall[i - 1][j] == 0:
                        finished = 0
                    elif mp[i - 1][j] == ".":
                        finished = 0

        for j in range(0,4):
            for i in range(2,-1,-1):
                if mp[i][j] != '.' and wall[i][j] == 0:
                    if mp[i + 1][j] == mp[i][j] and wall[i + 1][j] == 0:
                        finished = 0
                    elif mp[i + 1][j] == ".":
                        finished = 0

        for j in range(0,4):
            for i in range(1,4):
                if mp[j][i] != '.' and wall[j][i] == 0:
                    if mp[j][i - 1] == mp[j][i] and wall[j][i - 1] == 0:
                        finished = 0
                    elif mp[j][i - 1] == ".":
                        finished = 0

        for j in range(0,4):
            for i in range(2,-1,-1):
                if mp[j][i] != '.' and wall[j][i] == 0:
                    if mp[j][i + 1] == mp[j][i] and wall[j][i + 1] == 0:
                        finished = 0
                    elif mp[j][i + 1] == ".":
                        finished = 0

        if finished:
            break

        nxt = 1
        while(nxt == 1):
            event = keyboard.read_event()
            if event.event_type == keyboard.KEY_DOWN:
                if event.name == 'up':
                    for j in range(0,4):
                        move = 1
                        while move == 1:
                            move = 0
                            for i in range(1,4):
                                if mp[i][j] != '.' and wall[i][j] == 0:
                                    if mp[i - 1][j] == mp[i][j] and wall[i - 1][j] == 0:
                                        mp[i - 1][j] = 2 * mp[i][j]
                                        score += 2 * mp[i][j]
                                        mp[i][j] = '.'
                                        wall[i - 1][j] = 1
                                        nxt = 0
                                        move = 1
                                        os.system('cls')

                                    elif mp[i - 1][j] == ".":
                                        mp[i - 1][j] = mp[i][j]
                                        mp[i][j] = '.'
                                        move = 1
                                        nxt = 0
                                        os.system('cls')

                elif event.name == 'down':
                    for j in range(0,4):
                        move = 1
                        while move == 1:
                            move = 0
                            for i in range(2,-1,-1):
                                if mp[i][j] != '.' and wall[i][j] == 0:
                                    if mp[i + 1][j] == mp[i][j] and wall[i + 1][j] == 0:
                                        mp[i + 1][j] = 2 * mp[i][j]
                                        score += 2 * mp[i][j]
                                        mp[i][j] = '.'
                                        wall[i + 1][j] = 1
                                        move = 1
                                        nxt = 0
                                        os.system('cls')

                                    elif mp[i + 1][j] == ".":
                                        mp[i + 1][j] = mp[i][j]
                                        mp[i][j] = '.'
                                        move = 1
                                        nxt = 0
                                        os.system('cls')

                elif event.name == 'left':
                    for j in range(0,4):
                        move = 1
                        while move == 1:
                            move = 0
                            for i in range(1,4):
                                if mp[j][i] != '.' and wall[j][i] == 0:
                                    if mp[j][i - 1] == mp[j][i] and wall[j][i - 1] == 0:
                                        mp[j][i - 1] = 2 * mp[j][i]
                                        score += 2 * mp[j][i]
                                        mp[j][i] = '.'
                                        wall[j][i - 1] = 1
                                        nxt = 0
                                        move = 1
                                        os.system('cls')

                                    elif mp[j][i - 1] == ".":
                                        mp[j][i - 1] = mp[j][i]
                                        mp[j][i] = '.'
                                        move = 1
                                        nxt = 0
                                        os.system('cls')

                elif event.name == 'right':
                    for j in range(0,4):
                        move = 1
                        while move == 1:
                            move = 0
                            for i in range(2,-1,-1):
                                if mp[j][i] != '.' and wall[j][i] == 0:
                                    if mp[j][i + 1] == mp[j][i] and wall[j][i + 1] == 0:
                                        mp[j][i + 1] = 2 * mp[j][i]
                                        score += 2 * mp[j][i]
                                        mp[j][i] = '.'
                                        wall[j][i + 1] = 1
                                        nxt = 0
                                        move = 1
                                        os.system('cls')

                                    elif mp[j][i + 1] == ".":
                                        mp[j][i + 1] = mp[j][i]
                                        mp[j][i] = '.'
                                        move = 1
                                        nxt = 0
                                        os.system('cls')
            
    print("Game Over!")
                    



import random
from collections import deque
def cmd_minesweeping(a = 0, b = 0, c = 1):
    while a * b <= c:
        try:
            a, b, c = map(int,input("请输入行数，列数以及雷数：").split())
            if a * b <= c:
                print("雷太多了")
        except:
            print("你的输入有误，请再输一次")
            continue
    sizex = a
    sizey = b
    setbomb = c
    mp = [['.' for _ in range(sizey)] for _ in range(sizex)]
    bomb = [[0 for _ in range(sizey)] for _ in range(sizex)]
    # for i in mp:
    #     for j in i:
    #         print(j,end=" ")
    #     print()

    first_time = 1

    num_of_bomb = setbomb
    while(num_of_bomb > 0):
        try:
            x, y, z = map(int,input("请输入行，列，检测雷(1)或标记雷(2)：").split())
            x -= 1
            y -= 1
            if z != 1 and z != 2:
                print("你的输入有误，请再输一次")
                continue
            print()
            if mp[x][y] != '.':
                print("此格已被打开，请再输一次")
                continue

            if first_time == 1:
                labelnum = setbomb
                while(labelnum > 0):
                    xi = random.randint(0, sizex - 1)
                    yi = random.randint(0, sizey - 1)
                    if bomb[xi][yi] == 1 or (x == xi and y == yi):
                        continue
                    bomb[xi][yi] = 1
                    labelnum -= 1
                first_time = 0

            if (bomb[x][y] == 1 and z == 1) or (bomb[x][y] == 0 and z == 2):
                print("Game Over!")
                print("   ",end="")
                for i in range(1,sizey + 1):
                    print(i,end=" ")
                print()
                for i in range(sizex):
                    print(i + 1,end=" ")
                    if(i + 1 < 10):
                        print(" ",end="")
                    for j in range(sizey):
                        if bomb[i][j] == 1:
                            print('B',end=" ")
                        else:
                            print(mp[i][j],end=" ")
                    print()
                break

            print("   ",end="")
            for i in range(1,sizey + 1):
                print(i,end=" ")
            print()

            if bomb[x][y] == 1 and z == 2:
                mp[x][y] = 'B'
                num_of_bomb -= 1
                for i in range(sizex):
                    print(i + 1,end=" ")
                    if(i + 1 < 10):
                        print(" ",end="")
                    for j in range(sizey):
                        print(mp[i][j],end=" ")
                    print()
                print("剩余雷数：" + str(num_of_bomb))
                continue

            qx = deque()
            qy = deque()
            qx.append(x)
            qy.append(y)
            side = [-1,0,1]

            while len(qx) != 0:
                x = qx.popleft()
                y = qy.popleft()
                counts = 0
                for i in side:
                    if x + i < 0 or x + i > sizex - 1:
                        continue
                    for j in side:
                        if y + j < 0 or y + j > sizey - 1:
                            continue
                        if bomb[x + i][y + j] == 1:
                            counts += 1

                mp[x][y] = counts

                if counts == 0:
                    for i in side:
                        if x + i < 0 or x + i > sizex - 1:
                            continue
                        for j in side:
                            if y + j < 0 or y + j > sizey - 1:
                                continue
                            if bomb[x + i][y + j] != 1 and mp[x + i][y + j] == '.':
                                qx.append(x + i)
                                qy.append(y + j)
                

            
            for i in range(sizex):
                print(i + 1,end=" ")
                if(i + 1 < 10):
                    print(" ",end="")
                for j in range(sizey):
                    print(mp[i][j],end=" ")
                print()
            print("剩余雷数：" + str(num_of_bomb))

        except:
            print("你的输入有误，请再输一次")
            continue

    if num_of_bomb == 0:
        print("You win!")
        





from urllib.parse import unquote
from urllib.parse import quote

baseURL = 'https://molar-app-prod-v5.oss-cn-hangzhou.aliyuncs.com/'

def unqt(str):
    return unquote(str)

def qt(str):
    return quote(str)


import subprocess
from tqdm import tqdm
import multiprocessing
from pathlib import Path

def process(row, potree_converter_executable, output_folder_name):
    save_path = row.parent / output_folder_name

    save_path.mkdir(parents=True, exist_ok=True)
    command = [
        potree_converter_executable,
        row,
        "-o", save_path
        
    ]
    log = subprocess.run(command, capture_output=True)
    if log.returncode != 0:
        error_message = f"Error processing {row}: {log.stderr.decode()}\n"
        with open("conversion_errors.txt", "a") as error_file:
            error_file.write(error_message)

def las_2_octree(data_path, item_clip_name = "",
                potree_converter_url = r"D:\PotreeConverter_windows_x64\PotreeConverter.exe",
                output_folder_name = "combineLidar"):
    lasPath = []
    for item_clip in tqdm(Path(data_path).iterdir()):     # 每份las在clip目录下
        if item_clip_name == "" or item_clip.name == item_clip_name:
            lasPath.append(item_clip)

    print(f'待处理：{len(lasPath)}')

    with multiprocessing.Pool(15) as pool:
        for _ in tqdm(pool.imap_unordered(process, lasPath, potree_converter_url, output_folder_name), total=len(lasPath)):
            pass



import time
import random
import keyboard
import os
from collections import deque

def cmd_snake(x = 0, y = 0, z = 0):
    if x == 0 or y == 0:
        while 1:
            try:
                x, y = [int(i) for i in input("请输入地图长(>=8)和宽(>=8)：").split()]
                if x < 8 or y < 8:
                    continue
                break
            except:
                continue
    if z == 0:            
        while 1:
            try:
                z = int(input("请输入速度(1~5)："))
                if z < 1 or z > 5:
                    continue
                break
            except:
                continue
    mp = [['.' for _ in range(int(y))] for _ in range(int(x))]
    t = 0
    time_start = time.time()

    body = deque()
    head = (4,4)
    tail = (4,2)
    body.append(tail)
    body.append((4,3))
    body.append(head)
    mp[4][4] = 'O'
    mp[4][3] = 'o'
    mp[4][2] = 'o'

    while(1):
        xi = random.randint(0, int(x) - 1)
        yi = random.randint(0, int(y) - 1)
        if mp[xi][yi] == '.':
            mp[xi][yi] = '*'
            break

    direction = 'right'
    last = 'right'
    event = None
    score = 0
    while 1:
        os.system('cls')

        time_end = time.time()
        time_elapsed = time_end - time_start
        for i in mp:
            for j in i:
                print(j,end=" ")
            print()
        print("time:" + str(round(t, 2)) + " score:" + str(score))
        
        while time_elapsed < t:
            if keyboard.is_pressed('up') and last != 'down':
                direction = 'up'
            if keyboard.is_pressed('down') and last != 'up':
                direction = 'down'
            if keyboard.is_pressed('left') and last != 'right':
                direction = 'left'
            if keyboard.is_pressed('right') and last != 'left':
                direction = 'right'
            time_end = time.time()
            time_elapsed = time_end - time_start
        
        if direction == 'right':
            if mp[head[0]][(head[1] + 1) % int(y)] == '.' or tail == (head[0], (head[1] + 1) % int(y)):
                mp[tail[0]][tail[1]] = '.'
                body.popleft()
                tail = (body[0][0], body[0][1])
                mp[head[0]][head[1]] = 'o'
                mp[head[0]][(head[1] + 1) % int(y)] = 'O'
                head = (head[0], (head[1] + 1) % int(y))
                body.append(head)
                last = 'right'
            elif mp[head[0]][(head[1] + 1) % int(y)] == '*':
                score += z
                mp[head[0]][head[1]] = 'o'
                mp[head[0]][(head[1] + 1) % int(y)] = 'O'
                head = (head[0], (head[1] + 1) % int(y))
                body.append(head)
                while(1):
                    xi = random.randint(0, int(x) - 1)
                    yi = random.randint(0, int(y) - 1)
                    if mp[xi][yi] == '.':
                        mp[xi][yi] = '*'
                        break
                last = 'right'
            else:
                break

        if direction == 'left':
            lasth = head[1] - 1
            if head[1] - 1 < 0:
                lasth = int(y) - 1
            if mp[head[0]][lasth] == '.' or tail == (head[0], lasth):
                mp[tail[0]][tail[1]] = '.'
                body.popleft()
                tail = (body[0][0], body[0][1])
                mp[head[0]][head[1]] = 'o'
                mp[head[0]][lasth] = 'O'
                head = (head[0], lasth)
                body.append(head)
                last = 'left'
            elif mp[head[0]][lasth] == '*':
                score += z
                mp[head[0]][head[1]] = 'o'
                mp[head[0]][lasth] = 'O'
                head = (head[0], lasth)
                body.append(head)
                while(1):
                    xi = random.randint(0, int(x) - 1)
                    yi = random.randint(0, int(y) - 1)
                    if mp[xi][yi] == '.':
                        mp[xi][yi] = '*'
                        break
                last = 'left'
            else:
                break

        if direction == 'down':
            if mp[(head[0] + 1) % int(x)][head[1]] == '.' or tail == ((head[0] + 1) % int(x), head[1]):
                mp[tail[0]][tail[1]] = '.'
                body.popleft()
                tail = (body[0][0], body[0][1])
                mp[head[0]][head[1]] = 'o'
                mp[(head[0] + 1) % int(x)][head[1]]  = 'O'
                head = ((head[0] + 1) % int(x), head[1])
                body.append(head)
                last = 'down'
            elif mp[(head[0] + 1) % int(x)][head[1]]  == '*':
                score += z
                mp[head[0]][head[1]] = 'o'
                mp[(head[0] + 1) % int(x)][head[1]]  = 'O'
                head = ((head[0] + 1) % int(x), head[1])
                body.append(head)
                while(1):
                    xi = random.randint(0, int(x) - 1)
                    yi = random.randint(0, int(y) - 1)
                    if mp[xi][yi] == '.':
                        mp[xi][yi] = '*'
                        break
                last = 'down'
            else:
                break

        if direction == 'up':
            lasth = head[0] - 1
            if head[0] - 1 < 0:
                lasth = int(x) - 1
            if mp[lasth][head[1]] == '.' or tail == (lasth, head[1]):
                mp[tail[0]][tail[1]] = '.'
                body.popleft()
                tail = (body[0][0], body[0][1])
                mp[head[0]][head[1]] = 'o'
                mp[lasth][head[1]] = 'O'
                head = (lasth, head[1])
                body.append(head)
                last = 'up'
            elif mp[lasth][head[1]] == '*':
                score += z
                mp[head[0]][head[1]] = 'o'
                mp[lasth][head[1]] = 'O'
                head = (lasth, head[1])
                body.append(head)
                while(1):
                    xi = random.randint(0, int(x) - 1)
                    yi = random.randint(0, int(y) - 1)
                    if mp[xi][yi] == '.':
                        mp[xi][yi] = '*'
                        break
                last = 'up'
            else:
                break
            
        gap = 1 - z * 0.2
        if gap == 0:
            gap = 0.1

        t += gap

    print("Game Over!将在10秒后关闭……")
    time.sleep(10)


__version__ = "0.2.1"

from .json_op import *
from .file_op import *
from .image_op import *
from .camera_op import *
from .str_op import *
from .las_2_octree import las_2_octree
from .pcd_2_las import pcd_2_las
from . import game


import requests
from PIL import Image

def download_image(url, save_path):
    try:
        # 发送 GET 请求
        response = requests.get(url, stream=True)
        response.raise_for_status()  # 检查请求是否成功

        # 以二进制写入模式保存图片
        with open(save_path, 'wb') as file:
            for chunk in response.iter_content(1024):  # 分块写入，避免内存占用过大
                file.write(chunk)
        print(f"图片已保存至: {save_path}")
    except Exception as e:
        print(f"下载失败: {e}")

def crop_image(input_path, output_path, left, top, right, bottom):
    """
    裁剪图片并保存
    :param input_path: 输入图片路径
    :param output_path: 输出图片路径
    :param left: 左上角 x 坐标
    :param top: 左上角 y 坐标
    :param right: 右下角 x 坐标
    :param bottom: 右下角 y 坐标
    """
    with Image.open(input_path) as img:
        # 裁剪图片 (left, top, right, bottom)
        cropped_img = img.crop((left, top, right, bottom))
        cropped_img.save(output_path)
        print(f"图片已裁剪并保存至: {output_path}")


import json
import numpy as np
def camera_inverse_files(input_path, output_path):
    with open(input_path,'r',encoding='utf8')as fp:
        s = fp.read()
        data = json.loads(s)

    total = 0
    for i in data:
        A = np.array(i["extrinsic"])
        A_inv = np.linalg.inv(A)
        output = []
        t = 0
        for j in A_inv:
            output.append([])
            for k in j:
                output[t].append(float(k))
            t += 1
        
        data[total]['extrinsic'] = output
        total += 1

    json_str = json.dumps(data, indent=4, ensure_ascii=False)
    with open(output_path, "w+", encoding="utf-8") as file:
        file.write(json_str)

def camera_inverse_str(input_str):
    data = input_str

    total = 0
    for i in data:
        A = np.array(i["extrinsic"])
        A_inv = np.linalg.inv(A)
        output = []
        t = 0
        for j in A_inv:
            output.append([])
            for k in j:
                output[t].append(float(k))
            t += 1
        
        data[total]['extrinsic'] = output
        total += 1

    return data
                


# from lasWriter import write_las
# from pcdReader import read_pcd
from tqdm import tqdm
import numpy as np
from laspy import LasHeader, LasData
from multiprocessing import Pool, cpu_count
from pathlib import Path
import moore

def write_las(
    xyz: np.ndarray,
    intensity: np.ndarray = None,
    rgb: np.ndarray = None,
    path: str = "output.las",
) -> None:
    """LAS Format Floating-Point Storage Strategy

    Notes:
    - Real: real_x = int_x * scale + offset
    - Stored: int_x = (real_x - offset) / scale
    - Validation: normalized coordinates (xyz) must fall within [0.0, 1.0]

    References:
    - https://www.loc.gov/preservation/digital/formats/fdd/fdd000418.shtml
    - https://github.com/potree/PotreeConverter/blob/develop/Converter/src/chunker_countsort_laszip.cpp
    """
    # Create LAS Header
    header = LasHeader(version="1.4", point_format=2)

    scales = np.array([2**-10, 2**-10, 2**-10])  # 0.001 meter -> 1 mm
    mins = np.min(xyz, axis=0)
    maxs = np.max(xyz, axis=0)

    # Cube millimeter offset
    cube_size = maxs - mins
    mins_scaled = ((mins - mins) / scales).astype(np.int32).astype(np.float32)
    maxs_scaled = ((maxs - mins) / scales).astype(np.int32).astype(np.float32)
    normalized_mins = (mins_scaled * scales + mins - mins) / cube_size
    normalized_maxs = (maxs_scaled * scales + mins - mins) / cube_size
    if np.any(normalized_mins < 0) or np.any(normalized_maxs > 1):
        mins -= 1e-3
        maxs += 1e-3

    header.scales = scales
    header.mins = mins
    header.maxs = maxs

    # Create LAS Data
    las = LasData(header)

    las.x = xyz[:, 0]
    las.y = xyz[:, 1]
    las.z = xyz[:, 2]

    if intensity is not None:
        las.intensity = intensity.astype(np.uint16)

    if rgb is not None:
        r = rgb[:, 0].astype(np.uint16) * 256
        g = rgb[:, 1].astype(np.uint16) * 256
        b = rgb[:, 2].astype(np.uint16) * 256
        las.red, las.green, las.blue = r, g, b

    las.write(path)

def pcd_to_las(pcd_path: str, las_path: str) -> None:
    pc, _ = moore.read_pcd(pcd_path)  
    write_las(pc[:, :3], intensity=pc[:, 3], path=las_path)


def pcd_2_las(data_path, clip_dir_name = '') -> None:

    for clip_dir in tqdm(Path(data_path).iterdir()):
        if clip_dir_name == '' or clip_dir.name == clip_dir_name:
            path = clip_dir
            print(path.with_suffix(".las"))
            pcd_to_las(path, path.with_suffix(".las"))